/**
 * µP Steckdosen Logik Controller
 * Schaltet Steckdosen-Kombinationen nach Logik-Matrix
 * compile :g++ -Wall -o logic logic.cpp
 * Autor: Edgar Hermanns
 * (c) 2021 Edgar Hermanns
 */

#include <cstdint>
#include <cstdio>

enum Limits_e {
  LM_MaxDevPorts = 14
};

enum StateBits_e {
  SB_OnOff = 0 // the state is in bit 0: 0 == OFF, 1 == ON  
};

enum ConfigBits_e {
  CB_ManualOffOnly = 0 // the Manual-off-only bit in mConfig is in bit 0: 1 = only manual set-off ct, 0 = not only manual set-off
};

enum DevPortState_e {
  DPS_Off = 0,
  DPS_On = 1
};

enum ConfigAttr_e {
  CA_ManualOffOnly = 1  
};
  
struct Matrix_s {
  uint8_t mConfig;  // 8 bits for additional settings. bit 0 is defaultState  
  uint8_t mRefs[(LM_MaxDevPorts + 7)/8]; // bitwise references
  const char   mChoice;
  const char*  mName;
}; // struct Matrix_s


// -------------------- Labormatrix --------------------
Matrix_s PortMatrix[LM_MaxDevPorts] = {
  // first is top priority, i.e.: is switched first
  // bit g = switchable
  //     h = protected (no auto-off)
  //  abcdefgh
  { 0b00000000, { 0b00000000, 0b00000000 }, 's', "Switch" },           // switch ,         needs nothing else
  { 0b00000001, { 0b10010000, 0b00001000 }, 'l', "Laptop" },           // laptop,          needs switch, monitor 1, Laptop Powersupply
  { 0b00000001, { 0b10011100, 0b00000000 }, 'i', "Simbox" },           // Flightsimulator, needs switch, monitors 1,2,3
  { 0b00000000, { 0b00000000, 0b00000000 }, '1', "Monitor 1" },        // monitor,         needs nothing
  { 0b00000000, { 0b00000000, 0b00000000 }, '2', "Monitor 2" },        // monitor,         needs nothing
  { 0b00000000, { 0b00000000, 0b00000000 }, '3', "Monitor 3" },        // monitor,         needs nothing
  { 0b00000000, { 0b10000000, 0b00000000 }, 'd', "Drucker" },          // Drucker,         needs switch
  { 0b00000001, { 0b10000000, 0b00000000 }, 'a', "NAS A" },            // keysight,        needs switch
  { 0b00000001, { 0b00000001, 0b00000000 }, 'b', "NAS B" },            // NAS B,           needs NAS A
  { 0b00000000, { 0b10000000, 0b00000000 }, 'o', "Oscilloscope" },     // scope,           needs switch
  { 0b00000000, { 0b10000000, 0b00000000 }, 'k', "Keysight DMM" },     // keysight,        needs switch
  { 0b00000000, { 0b10000000, 0b00000000 }, 'g', "Signal generator" }, // keysight,        needs switch
  { 0b00000000, { 0b00000000, 0b00000000 }, 'y', "Laptop PwrSupply" }, // keysight,        needs switch
  { 0b00000000, { 0b00000000, 0b00000000 }, 'p', "PSU" },              // keysight,        needs nothing
};

class DevPort {
  // device port class
public:
  DevPort (uint8_t aNumber, DevPort* aDevPorts);
  void dump();
  void aquire();
  void release(bool aByUser = false, bool aForce = false);
  void toggle(bool aByUser = false, bool aForce = false);
  bool isOn();
  bool isOff();
  bool isManualOffOnly();
  bool needs(uint8_t aDevPortNumber);
  uint8_t getInUseBy(DevPort* aDevPorts);
  
private:
  uint8_t mNumber;
  uint8_t mState;
  uint8_t getNumber();
  void setState(StateBits_e aBit, DevPortState_e aState);
}; // class DevPort

DevPort DevPorts[LM_MaxDevPorts] {
                  { 0, DevPorts},
                  { 1, DevPorts},
                  { 2, DevPorts},
                  { 3, DevPorts},
                  { 4, DevPorts},
                  { 5, DevPorts},
                  { 6, DevPorts},
                  { 7, DevPorts},
                  { 8, DevPorts},
                  { 9, DevPorts},
                  {10, DevPorts},
                  {11, DevPorts},
                  {12, DevPorts},
                  {13, DevPorts},
                    }; // devports

DevPort::DevPort (uint8_t aNumber, DevPort* aDevPorts)
  : mNumber(aNumber)
  , mState(0) {
  // ---
} // DevPort ctor

void DevPort::aquire() {
  // walk through matrix and check which device shall be aquired
  setState(SB_OnOff, DPS_On);
  bool hadChange = false;
  do {
    hadChange = false;
    for (uint8_t i = 0; i < LM_MaxDevPorts; ++i) {

      // myself ?
      if (i == mNumber)
        continue;    

      if (DevPorts[mNumber].needs(i)) {
        if (DevPorts[i].isOff()) {
          DevPorts[i].aquire();
          hadChange = true;
        } // devport OFF
      } // if ports needs ...
    } // for i..8
  } while (hadChange);
  
} // aquire

void DevPort::release(bool aByUser, bool aForce) {
  printf("DevPort %d is in use by %d devices\n", mNumber, getInUseBy(DevPorts));

  if (!aForce)
    if ((!aByUser && isManualOffOnly()) ||getInUseBy(DevPorts) > 0)
      return;
  
  setState(SB_OnOff, DPS_Off);

  // on a forced OFF, only switch this single device.
  if (aForce)
    return;
  
  for (uint8_t i = 0; i < LM_MaxDevPorts; ++i) {
    if (i == mNumber || DevPorts[i].isManualOffOnly())
      continue;
    if (DevPorts[i].getInUseBy(DevPorts) == 0)
      DevPorts[i].setState(SB_OnOff, DPS_Off);
  }  // i .. LM_MaxDevPorts
} // release

bool DevPort::needs(uint8_t aDevPortNumber) {
  // tbd: index mrefs is fixed to '0'. shall be calculated if more than 8 outlets exist
  uint8_t refIdx = aDevPortNumber/8;
  uint8_t bit = aDevPortNumber % 8;
  return (PortMatrix[mNumber].mRefs[refIdx] & (1<<(7 - bit))) > 0;
}

bool DevPort::isOn() {
  return (mState & (1 << SB_OnOff)) != 0;
}

bool DevPort::isOff() {
  return (mState & (1 << SB_OnOff)) == 0;
}

bool DevPort::isManualOffOnly () {
  return (PortMatrix[mNumber].mConfig & (1 << CB_ManualOffOnly)) != 0;
}

void DevPort::setState(StateBits_e aBit, DevPortState_e aState) {
  mState &= ~(1 << aBit);
  mState |=  (aState << aBit);
} // DevPort::setState

uint8_t DevPort::getInUseBy(DevPort* aDevPorts) {  
  // check which device currently uses DevPorts[aCheckNumber]
  uint8_t useCount = 0;
  for (uint8_t i = 0; i < LM_MaxDevPorts; ++i) {
    if (i == mNumber) // myself
      continue;
    if (DevPorts[i].isOff())
      continue;
    
    // check, if device[i] needs device[aCheckNumber]
    if (DevPorts[i].needs(mNumber))
      ++useCount;
  }
  return useCount;
} // getInUseBy

void DevPort::toggle(bool aByUser, bool aForce) {
  if (isOn())
    release(aByUser, aForce);
  else
    aquire();
}

void dumpMatrix(StateBits_e aBit) {
  // dumps ALL outlet state bits 'aBit'
  printf ("Matrix bit[%d]: [", aBit);
  char comma[2] = "";
  for (uint8_t i = 0; i < 8; ++i) {
    printf("%s%d(%d)", comma,(DevPorts[i].isOn() ? 1:0), DevPorts[i].getInUseBy(DevPorts));
    comma[0] = ',';
  }  
  printf ("]\n");
}

int main() {  
  int ch = 0;
  do {
    printf (
      "\n\nEnter letter to toggle device on/off. Twice the letter is 'force'. ...\n");
    
    for (uint8_t i = 0; i < LM_MaxDevPorts; ++i) {
      printf ("'%c' %-20.20s [%s]\n", PortMatrix[i].mChoice, PortMatrix[i].mName, (DevPorts[i].isOn()) ? "ON" : "OFF");
    }

    printf(
      "\n"
      "'q' to quit\n\n"
      "-----------\n"
      "Input: ");
    
    bool force = false;
    ch = getchar();

    // twice ?
    if (getchar() == ch) {
      force = true;
      getchar(); // flush newline
    } // getchat == ch

    if (force)
      printf ("FORCED action\n");

    // search entry and toggle
    for (uint8_t i = 0; i < LM_MaxDevPorts; ++i) {
      if (ch == PortMatrix[i].mChoice) {
        DevPorts[i].toggle(true, force);  break;
      } // if mChoice ...
    } // for i..8
    
    dumpMatrix(SB_OnOff);           
  }  while (ch != 'q');
} // main
